#!/usr/bin/env python

__copyright__ = "Copyright (c) 2020-2025 Alex Laird"
__license__ = "MIT"

import click
import platform

from hookee import HookeeManager, __version__, pluginmanager


@click.group(invoke_without_command=True)
@click.pass_context
@click.option("--port", type=int, help="[server] The local port for the webserver and ngrok tunnel.")
@click.option('--default-route', type=str, help="[server] The URI regex to map to the default webhook.")
@click.option('--no-tunnel', is_flag=True, default=False, help="Do not open an ngrok tunnel.")
@click.option("--response", type=str,
              help="[server] Data to set for the response, overriding all body data from plugins and "
                   "`--response-script`.")
@click.option("--content-type", type=str,
              help="[server] The \"Content-Type\" header to set when response body data is given with `--response`")
@click.option("--request-script", type=click.Path(exists=True),
              help="[server] A Python script whose `run(request)` method will be called by the default `/webhook` "
                   "after all request plugins have run.")
@click.option("--response-script", type=click.Path(exists=True),
              help="[server] A Python script whose `run(request, response)` method will be called by the default "
                   "`/webhook` after all response plugins have run.")
@click.option("--tunnel-name", help="[tunnel] The name to use for the ngrok tunnel.")
@click.option("--subdomain", help="[tunnel] The subdomain to use for ngrok endpoints.")
@click.option("--region", type=click.Choice(["us", "eu", "ap", "au", "sa", "jp", "in"]),
              help="The region to use for ngrok endpoints.")
@click.option("--hostname", help="[tunnel] The domain to use for ngrok endpoints.")
@click.option("--auth", help="[tunnel] The basic auth to use for ngrok endpoints.")
@click.option("--host-header", help="[tunnel] The \"Host\" header value to use for ngrok endpoints.")
@click.option("--auth-token", help="[tunnel] A valid ngrok auth token.")
@click.option("--api-key", help="[tunnel] A valid ngrok API key.")
@click.option("--plugins-dir", type=click.Path(exists=True), help="The directory to scan for custom hookee plugins.")
@click.option("--plugins", multiple=True, help="A list of hookee plugins to use.")
@click.option('--version', is_flag=True, default=False, help="Display version information.")
def hookee(ctx, **kwargs):
    """
    hookee is a utility that provides command line webhooks, on demand! Bind port to intercept requests,
    dump request data to the console, process requests and responses, customize response data, and configure hookee and
    its routes further in any number of ways through custom plugins.

    hookee can be started by using `hookee start` or simply hookee.

    If options are given, they override the default values derived from the config file.

    hookee documentation can be found at https://hookee.readthedocs.io.
    """
    if kwargs["version"]:
        click.echo(f"hookee/{__version__} Python/{platform.python_version()}")
        ctx.exit(0)

    ctx.ensure_object(dict)
    for key, value in kwargs.items():
        if value:
            ctx.obj[key] = value

    if kwargs.get("subdomain") and kwargs.get("hostname"):
        ctx.fail("Can't give both --subdomain and --hostname.")

    hookee_manager = HookeeManager(load_plugins=ctx.invoked_subcommand not in [enable_plugin.name,
                                                                               disable_plugin.name,
                                                                               available_plugins.name,
                                                                               enabled_plugins.name])
    ctx.obj["hookee_manager"] = hookee_manager

    if ctx.invoked_subcommand is None:
        hookee_manager.run()


@hookee.command()
@click.pass_context
def start(ctx):
    """
    Start hookee.
    """
    hookee_manager = ctx.obj["hookee_manager"]

    hookee_manager.run()


@hookee.command(
    short_help="Update the default value for a config. Any passable arg to hookee can also be given here to set its "
               "default in the config so it doesn't need to be passed to the hookee each time."
)
@click.pass_context
@click.argument("key")
@click.argument("value")
def update_config(ctx, key, value):
    """
    Update the default value for a config. Any passable arg to hookee can also be given here to set its
    default in the config so it doesn't need to be passed to the hookee each time.
    """
    hookee_manager = ctx.obj["hookee_manager"]

    if key == "plugins":
        ctx.fail("Enable and disable plugins through the `enable-plugin` and `disable-plugin` commands.")
    if "-" in key:
        key = key.replace("-", "_")

    if value.isdigit():
        value = int(value)

    try:
        hookee_manager.config.set(key, value)

        hookee_manager.print_util.print_config_update(
            f"The default value for \"{key}\" has been updated in the config.")
    except KeyError:
        ctx.fail(f"No such key exists in the config: {key}")


@hookee.command()
@click.pass_context
@click.argument("plugin")
def enable_plugin(ctx, plugin):
    """
    Enable the given plugin by default.
    """
    hookee_manager = ctx.obj["hookee_manager"]

    hookee_manager.plugin_manager.get_plugin(plugin)

    hookee_manager.config.append("plugins", plugin)

    hookee_manager.print_util.print_config_update(f"Plugin \"{plugin}\" has been enabled.")


@hookee.command()
@click.pass_context
@click.argument("plugin")
def disable_plugin(ctx, plugin):
    """
    Disable the given plugin by default.
    """
    hookee_manager = ctx.obj["hookee_manager"]

    if plugin in pluginmanager.REQUIRED_PLUGINS:
        ctx.fail(f"Can't disable the plugin \"{plugin}\".")

    hookee_manager.config.remove("plugins", plugin)

    hookee_manager.print_util.print_config_update(f"Plugin \"{plugin}\" is disabled.")


@hookee.command()
@click.pass_context
def available_plugins(ctx):
    """
    List all available plugins.
    """
    hookee_manager = ctx.obj["hookee_manager"]

    plugins = hookee_manager.plugin_manager.available_plugins()

    hookee_manager.print_util.print_open_header("Available Plugins")

    for plugin_name in plugins:
        hookee_manager.print_util.print_basic(f" * {plugin_name}")

        try:
            plugin = hookee_manager.plugin_manager.get_plugin(plugin_name)
            if plugin.description:
                hookee_manager.print_util.print_basic(f"   Description: {plugin.description}")
        except Exception as e:
            hookee_manager.print_util.print_basic(f"   Error: {e}")

    hookee_manager.print_util.print_close_header()
    hookee_manager.print_util.print_basic()


@hookee.command()
@click.pass_context
def enabled_plugins(ctx):
    """
    List all enabled plugins.
    """
    hookee_manager = ctx.obj["hookee_manager"]

    plugins = hookee_manager.plugin_manager.enabled_plugins()

    hookee_manager.print_util.print_open_header("Enabled Plugins (Order of Execution)")

    for plugin_name in plugins:
        hookee_manager.print_util.print_basic(f" * {plugin_name}")

        try:
            plugin = hookee_manager.plugin_manager.get_plugin(plugin_name)
            if plugin.description:
                hookee_manager.print_util.print_basic(f"   Description: {plugin.description}")
        except Exception as e:
            hookee_manager.print_util.print_basic(f"   Error: {e}")

    hookee_manager.print_util.print_close_header()
    hookee_manager.print_util.print_basic()


@hookee.command()
@click.pass_context
@click.argument("token")
def authtoken(ctx, token):
    """
    Save the auth token to the config.
    """
    hookee_manager = ctx.obj["hookee_manager"]

    hookee_manager.config.set("auth_token", token)

    hookee_manager.print_util.print_config_update("The auth token has been set in the config.")


@hookee.command()
@click.pass_context
@click.argument("key")
def apikey(ctx, key):
    """
    Save the API key to the config.
    """
    hookee_manager = ctx.obj["hookee_manager"]

    hookee_manager.config.set("api_key", key)

    hookee_manager.print_util.print_config_update("The API key has been set in the config.")


if __name__ == "__main__":
    hookee(obj={})



__copyright__ = "Copyright (c) 2020-2024 Alex Laird"
__license__ = "MIT"

from datetime import datetime

from hookee.pluginmanager import REQUEST_PLUGIN
from hookee.util import PrintUtil  # noqa: F401

plugin_type = REQUEST_PLUGIN
description = "Print the timestamp along with the request URL and method being invoked."

print_util = None  # type: PrintUtil


def setup(hookee_manager):
    global print_util

    print_util = hookee_manager.print_util


def run(request):
    now = datetime.now()

    timestamp = now.strftime("%m-%d-%Y %I:%M:%S %p")

    print_util.print_basic(
        "[{timestamp}] \"{method} {url} {protocol}\"".format(timestamp=timestamp,
                                                             method=request.method,
                                                             url=request.base_url,
                                                             protocol=request.environ["SERVER_PROTOCOL"]),
        color=print_util.request_color, bold=True)
    print_util.print_basic()

    return request



__copyright__ = "Copyright (c) 2020-2024 Alex Laird"
__license__ = "MIT"

import os
import unittest

import requests

from hookee import HookeeManager
from hookee.conf import Config
from tests.testcase import HookeeTestCase


class TestHookeeManagerEdges(HookeeTestCase):
    def test_not_click_logging(self):
        self.assertFalse(self.config.click_logging)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_hookee_manager(self):
        # GIVEN
        config = Config(port=8001)
        hookee_manager = HookeeManager(config=config)

        # THEN
        self.assertIsNone(hookee_manager.server._thread)
        self.assertIsNone(hookee_manager.tunnel._thread)

        # WHEN
        hookee_manager._init_server_and_tunnel()

        # THEN
        self.assertIsNotNone(hookee_manager.server._thread)
        self.assertIsNotNone(hookee_manager.tunnel._thread)

    @unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
    def test_custom_response(self):
        # GIVEN
        response_body = "<Response>Ok</Response>"
        config = Config(response=response_body, auth_token=os.environ["NGROK_AUTHTOKEN"], port=8002)
        hookee_manager = HookeeManager(config=config)
        hookee_manager._init_server_and_tunnel()

        webhook_url = f"{hookee_manager.tunnel.public_url}/webhook"

        # WHEN
        response = requests.get(webhook_url)

        # THEN
        self.assertEqual(response.content.decode("utf-8"), response_body)



__copyright__ = "Copyright (c) 2020-2024 Alex Laird"
__license__ = "MIT"

from hookee.pluginmanager import REQUEST_PLUGIN
from hookee.util import PrintUtil  # noqa: F401

plugin_type = REQUEST_PLUGIN
description = "Print the `request`'s headers, if defined."

print_util = None  # type: PrintUtil


def setup(hookee_manager):
    global print_util

    print_util = hookee_manager.print_util


def run(request):
    if request.headers and "X-Forwarded-For" in request.headers:
        print_util.print_basic(f"Client IP: {request.headers.get('X-Forwarded-For')}",
                               color=print_util.request_color)
    if request.headers:
        print_util.print_dict("Headers", dict(request.headers), color=print_util.request_color)

    return request



__copyright__ = "Copyright (c) 2020-2024 Alex Laird"
__license__ = "MIT"

import os
import unittest

import requests

from tests.managedtestcase import ManagedTestCase


@unittest.skipIf(not os.environ.get("NGROK_AUTHTOKEN"), "NGROK_AUTHTOKEN environment variable not set")
class TestHookeeManager(ManagedTestCase):
    def test_http_get_query_params(self):
        # GIVEN
        params = {"param_1": ["param_value_1"]}

        # WHEN
        with self.captured_output() as (out, err):
            response = requests.get(self.webhook_url, params=params)

        self.assertEqual(response.status_code, 200)
        self.assertIn("""Query Params: {
    "param_1": "param_value_1"
}""", out.getvalue())

    def test_http_post_form_data(self):
        # GIVEN
        data = {"form_data_1": "form_data_value_1"}

        # WHEN
        with self.captured_output() as (out, err):
            response = requests.post(self.webhook_url, data=data)

        self.assertEqual(response.status_code, 200)
        self.assertIn("\"Content-Type\": \"application/x-www-form-urlencoded\"", out.getvalue())
        self.assertEqual(response.headers.get("Content-Type"), "application/json")
        self.assertIn("""Body: {
    "form_data_1": "form_data_value_1"
}""", out.getvalue())
        self.assertEqual(response.json(), data)

    def test_http_post_json_data(self):
        # GIVEN
        headers = {"Content-Type": "application/json"}
        data = {"json_data_1": "json_data_value_1"}

        # WHEN
        with self.captured_output() as (out, err):
            response = requests.post(self.webhook_url, headers=headers, json=data)

        self.assertEqual(response.status_code, 200)
        self.assertIn("\"Content-Type\": \"application/json\"", out.getvalue())
        self.assertIn("""Body: {
    "json_data_1": "json_data_value_1"
}""", out.getvalue())
        self.assertEqual(response.headers.get("Content-Type"), "application/json")
        self.assertEqual(response.json(), data)



__copyright__ = "Copyright (c) 2020-2024 Alex Laird"
__license__ = "MIT"

import os
import re
import signal

from flask import Blueprint, abort, request

from hookee.conf import Config  # noqa: F401
from hookee.pluginmanager import BLUEPRINT_PLUGIN, PluginManager  # noqa: F401
from hookee.util import PrintUtil  # noqa: F401

blueprint = Blueprint("default", __name__)
plugin_type = BLUEPRINT_PLUGIN
description = ("Mount required management endpoints along with the default `/webhook` endpoint that processes "
               "incoming requests with enabled plugins.")

plugin_manager = None  # type: PluginManager
print_util = None  # type: PrintUtil
config = None  # type: Config


def setup(hookee_manager):
    global plugin_manager, print_util, config

    plugin_manager = hookee_manager.plugin_manager
    print_util = hookee_manager.print_util
    config = hookee_manager.config


@blueprint.route('/<path:uri>',
                 methods=["GET", "HEAD", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "TRACE", "CONNECT"])
def webhook(uri=None):
    uri = uri.rstrip("/")
    if (not re.compile(config.get("default_route_methods")).match(request.method) or
            not re.compile(config.get("default_route")).match(f"/{uri}")):
        abort(404)

    print_util.print_close_header(delimiter="=", color=print_util.request_color)

    print_util.print_open_header("Request", delimiter="-", color=print_util.request_color)

    plugin_manager.run_request_plugins(request)

    print_util.print_open_header("Response", color=print_util.request_color)

    response = plugin_manager.run_response_plugins(request)

    print_util.print_close_header("=", color=print_util.request_color)
    print_util.print_basic()

    return response


@blueprint.route("/status")
def status():
    return "", 200


@blueprint.route("/shutdown", methods=["POST"])
def shutdown():
    if "werkzeug.server.shutdown" in request.environ:
        request.environ.get("werkzeug.server.shutdown")()
    else:
        # Windows does not provide SIGKILL, go with SIGTERM then
        sig = getattr(signal, "SIGKILL", signal.SIGTERM)
        os.kill(os.getpid(), sig)

    return "", 204
